/*
 * File sineqm.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL),
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
   Module: sineqm.c  - example TINE sine generator server, which illustrates
           the basic server API
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "sineqm.h"

#if defined(UNIX) || defined(WIN32)
# define random(r) (r != 0 ? rand()%(r) : 0)
#endif

#define PRP_SINE      1
#define PRP_AMPLITUDE 2
#define PRP_FREQUENCY 3
#define PRP_PHASE     4
#define PRP_NOISE     5
#define PRP_INFO      6

#define NUM_DEVICES 10
#define NUM_VALUES (1024*8)

/*
   The code section below illustrates how one registers a structure
   to be exported.
*/

typedef struct
{
  float amplitude;
  float frequency;
  float noise;
  float phase;
  int   numberCalls;
  char  description[64];
} SineInfo;

#define quit(i) { printf("Register struct: out of memory\n"); exit(i); }
void registerStructs(void)
{
  /* this must follow the order of the structure explicitly! */
  if (addFieldToStruct("SineInfo",OFFSETIN(SineInfo,amplitude),1,CF_FLOAT,"amplitude")) quit(1);
  if (addFieldToStruct("SineInfo",OFFSETIN(SineInfo,frequency),1,CF_FLOAT,"frequency")) quit(1);
  if (addFieldToStruct("SineInfo",OFFSETIN(SineInfo,noise),1,CF_FLOAT,"noise")) quit(1);
  if (addFieldToStruct("SineInfo",OFFSETIN(SineInfo,phase),1,CF_FLOAT,"phase")) quit(1);
  if (addFieldToStruct("SineInfo",OFFSETIN(SineInfo,numberCalls),1,CF_LONG,"numberCalls")) quit(1);
  if (addFieldToStruct("SineInfo",OFFSETIN(SineInfo,description),64,CF_TEXT,"description")) quit(1);
  /* terminate the structure definition like this! */
  if (sealTaggedStruct("SineInfo",sizeof(SineInfo),NUM_DEVICES)) quit(1);
}

SineInfo sineInfoTable[NUM_DEVICES];
float sinbuf[NUM_DEVICES][NUM_VALUES];

void updateSine(int devnr)
{ /* used internally -> don't bother with bounds checking */
  int i, k;
  float *target = sinbuf[devnr];
  float f = sineInfoTable[devnr].frequency;
  int n = (int)sineInfoTable[devnr].noise;
  int p = (int)sineInfoTable[devnr].phase;
  float a = sineInfoTable[devnr].amplitude;

  for (i=0; i<NUM_VALUES; i++)
  {
    k = (i + p) % NUM_VALUES;
    target[k] = random(n) + (float)(sin(f*i*6.2832/(NUM_VALUES/8))*a);
  }
  return;
}
char *sinname[NUM_DEVICES] = {
  "SineGen0","SineGen1","SineGen2","SineGen3","SineGen4",
  "SineGen5","SineGen6","SineGen7","SineGen8","SineGen9" };

void sininit(void)
{
  int i;
  DTYPE dout, din;

  /* some initial values for all sine devices */
  for (i=0; i<NUM_DEVICES; i++)
  {
    sineInfoTable[i].amplitude = 256.0;
    sineInfoTable[i].frequency = 1.0;
    sineInfoTable[i].phase = 0.0;
    sineInfoTable[i].noise = 30.0;
    sprintf(sineInfoTable[i].description,"Sine Generator %d at your disposal",i);
  }
  registerStructs();

  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 8192;
  dout.dTag[0] = 0;
  din.dFormat = CF_NULL;
  din.dArrayLength = 0;
  din.dTag[0] = 0;
  RegisterPropertyInformation(SINEQM_TAG,"Sine",&dout,&din,CA_READ,AT_SPECTRUM,8192,"[-1000:1000 V][0:1000 ms]Sine Curve",PRP_SINE,NULL);

  dout.dArrayLength = 10;
  din.dFormat = CF_FLOAT;
  din.dArrayLength = 1;
  RegisterPropertyInformation(SINEQM_TAG,"Amplitude",&dout,&din,CA_READ,AT_CHANNEL,10,"[1:1000 V]Sine Curve Amplitude",PRP_AMPLITUDE,NULL);
  RegisterPropertyInformation(SINEQM_TAG,"Frequency",&dout,&din,CA_READ,AT_CHANNEL,10,"[1:60]Sine Curve Frequency",PRP_FREQUENCY,NULL);
  RegisterPropertyInformation(SINEQM_TAG,"Phase",&dout,&din,CA_READ,AT_CHANNEL,10,"[0:512]Sine Curve Phase",PRP_PHASE,NULL);
  RegisterPropertyInformation(SINEQM_TAG,"Noise",&dout,&din,CA_READ,AT_CHANNEL,10,"[0:100 V]Sine Curve Noise Level",PRP_NOISE,NULL);

  dout.dFormat = CF_STRUCT;
  strncpy(dout.dTag,"SineInfo",8);
  din.dFormat = CF_STRUCT;
  strncpy(din.dTag,"SineInfo",8);
  RegisterPropertyInformation(SINEQM_TAG,"SineInfo",&dout,&din,CA_READ,AT_UNKNOWN,10,"Sine Generator Information",PRP_INFO,NULL);

  /* register device names, bind to device number: */
  for (i=0; i<NUM_DEVICES; i++) RegisterDeviceName(SINEQM_TAG,sinname[i],i);
}

void sinbkg(void)
{
  /* static int ptr = 0; */
  /* static int cnt = 0; */
  int i;
  /* do IO or something useful here */
  for (i=0; i<NUM_DEVICES; i++)
  {
    updateSine(i);
  }
}
int sineqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int devnr,prpid,i,cc;
  float mcarray[NUM_DEVICES], fval;
  SineInfo *sinf;

  devnr = GetDeviceNumber(SINEQM_TAG,devName);
  if (devnr < 0 || devnr >= NUM_DEVICES) return illegal_equipment_number;

  prpid = GetPropertyId(SINEQM_TAG,devProperty);

  switch (prpid)
  {
    case PRP_SINE:
        if (access&CA_WRITE) return illegal_read_write;
        if ((cc=putValuesFromFloat(dout,sinbuf[devnr],NUM_VALUES)) != 0) return cc;
        sineInfoTable[devnr].numberCalls++;
        return 0;
    case PRP_AMPLITUDE:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (fval < 1 || fval > 1000) return out_of_range;
          sineInfoTable[devnr].amplitude = fval;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++)
            mcarray[i] = sineInfoTable[i].amplitude;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_FREQUENCY:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (fval < 1 || fval > 100) return out_of_range;
          sineInfoTable[devnr].frequency = fval;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++)
            mcarray[i] = sineInfoTable[i].frequency;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_PHASE:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (fval < 0 || fval > 512) return out_of_range;
          sineInfoTable[devnr].phase = fval;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++)
            mcarray[i] = sineInfoTable[i].phase;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_NOISE:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (fval < 0 || fval > 100) return out_of_range;
          sineInfoTable[devnr].noise = fval;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++)
            mcarray[i] = sineInfoTable[i].noise;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_INFO:
        if (din->dArrayLength)
        { /* allow an atomic set of all relevant values this way : */
          if (din->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(din->dTag,"SineInfo",8)) return invalid_structure_tag;
          sinf = (SineInfo *)(din->data.vptr);
          if (sinf->amplitude < 1 || sinf->amplitude > 1000) return out_of_range;
          if (sinf->frequency < 1 || sinf->frequency > 100) return out_of_range;
          if (sinf->phase < 0 || sinf->phase > 512) return out_of_range;
          if (sinf->noise < 0 || sinf->noise > 100) return out_of_range;
          sineInfoTable[devnr] = *sinf;
        }
        if (dout->dArrayLength)
        {
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(dout->dTag,"SineInfo",8)) return invalid_structure_tag;
          if (dout->dArrayLength > NUM_DEVICES) dout->dArrayLength = NUM_DEVICES;
          sinf = (SineInfo *)(dout->data.vptr);
          for (i=0; i<(int)dout->dArrayLength; i++)
          {
            sinf[i] = sineInfoTable[(i+devnr)%NUM_DEVICES];
          }
        }
        return 0;
    default:
        return illegal_property;
  }
}
#define SINEGEN_FECNAME "SINEFEC"
#define SINEGEN_FECPORT 19
#define SINEGEN_CONTEXT "TEST"
#define SINEGEN_SUBSYS  "TEST"
void PreSystemInit(void)
{
# ifdef FIND_ENS_AND_EXPORTNAME
  char exp[20],ctx[20],fec[20];
  int port;
  FindNameServerOnNetwork();
  GetMyServerAddress(SINEQM_TAG,exp,ctx,fec,&port);
  RegisterFecInformation(fec,"TEST",ctx,"Test Sine Curve",
    "My Office","none","me",(UINT16)port);
# else
  RegisterFecInformation(SINEGEN_FECNAME,SINEGEN_CONTEXT,SINEGEN_SUBSYS,
    "Test Sine Curve","My Office","none","me",SINEGEN_FECPORT);
# endif
  SetCycleDelay(0);
}
void PostSystemInit(void)
{
  RegisterEquipmentModule(SINEQM_EXPORT,SINEQM_TAG,NUM_DEVICES,sineqm,sininit,sinbkg,200,NULL);
  /* don't let people poll faster than 20 Hz */
  MinPollingRate = MaxPollingRate = 50;

}
